# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Dict, List

from adi.context_manager import context_manager
from adi.rx_tx import rx_tx
from adi.sync_start import sync_start


def _map_to_dict(paths, ch):
    fddc, cddc, adc = ch.attrs["label"].value.split("->")
    if adc not in paths.keys():
        paths[adc] = {}
    if cddc not in paths[adc].keys():
        paths[adc][cddc] = {}
    if fddc not in paths[adc][cddc].keys():
        paths[adc][cddc][fddc] = {"channels": [ch._id]}
    else:
        paths[adc][cddc][fddc]["channels"].append(ch._id)
    return paths


def _sortconv(chans_names, noq=False, dds=False):
    tmpI = filter(lambda k: "_i" in k, chans_names)
    tmpQ = filter(lambda k: "_q" in k, chans_names)

    def ignoreadc(w):
        return int(w[len("voltage") : w.find("_")])

    def ignorealt(w):
        return int(w[len("altvoltage") :])

    chans_names_out = []
    if dds:
        filt = ignorealt
        tmpI = chans_names
        noq = True
    else:
        filt = ignoreadc

    tmpI = sorted(tmpI, key=filt)
    tmpQ = sorted(tmpQ, key=filt)
    for i in range(len(tmpI)):
        chans_names_out.append(tmpI[i])
        if not noq:
            chans_names_out.append(tmpQ[i])

    return chans_names_out


class ad9081(rx_tx, context_manager, sync_start):
    """AD9081 Mixed-Signal Front End (MxFE)"""

    _complex_data = True
    _rx_channel_names: List[str] = []
    _tx_channel_names: List[str] = []
    _tx_control_channel_names: List[str] = []
    _rx_coarse_ddc_channel_names: List[str] = []
    _tx_coarse_duc_channel_names: List[str] = []
    _rx_fine_ddc_channel_names: List[str] = []
    _tx_fine_duc_channel_names: List[str] = []
    _dds_channel_names: List[str] = []
    _device_name = ""

    _rx_attr_only_channel_names: List[str] = []
    _tx_attr_only_channel_names: List[str] = []

    _path_map: Dict[str, Dict[str, Dict[str, List[str]]]] = {}

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)
        # Default device for attribute writes
        self._ctrl = self._ctx.find_device("axi-ad9081-rx-hpc")
        # Devices with buffers
        self._rxadc = self._ctx.find_device("axi-ad9081-rx-hpc")
        self._txdac = self._ctx.find_device("axi-ad9081-tx-hpc")

        # Get DDC and DUC mappings
        paths = {}

        for ch in self._rxadc.channels:
            if "label" in ch.attrs:
                paths = _map_to_dict(paths, ch)
        self._path_map = paths

        # Get data + DDS channels
        for ch in self._rxadc.channels:
            if ch.scan_element and not ch.output:
                self._rx_channel_names.append(ch._id)
        for ch in self._txdac.channels:
            if ch.scan_element:
                self._tx_channel_names.append(ch._id)
            else:
                self._dds_channel_names.append(ch._id)

        # Sort channel names
        self._rx_channel_names = _sortconv(self._rx_channel_names)
        self._tx_channel_names = _sortconv(self._tx_channel_names)
        self._dds_channel_names = _sortconv(self._dds_channel_names, dds=True)

        # Map unique attributes to channel properties
        self._rx_fine_ddc_channel_names = []
        self._rx_coarse_ddc_channel_names = []
        self._tx_fine_duc_channel_names = []
        self._tx_coarse_duc_channel_names = []
        for converter in paths:
            for cdc in paths[converter]:
                channels = []
                for fdc in paths[converter][cdc]:
                    channels += paths[converter][cdc][fdc]["channels"]
                channels = [name for name in channels if "_i" in name]
                if "ADC" in converter:
                    self._rx_coarse_ddc_channel_names.append(channels[0])
                    self._rx_fine_ddc_channel_names += channels
                else:
                    self._tx_coarse_duc_channel_names.append(channels[0])
                    self._tx_fine_duc_channel_names += channels

        rx_tx.__init__(self)
        sync_start.__init__(self)
        self.rx_buffer_size = 2 ** 16

    def _get_iio_attr_str_single(self, channel_name, attr, output):
        # This is overridden by subclasses
        return self._get_iio_attr_str(channel_name, attr, output)

    def _set_iio_attr_str_single(self, channel_name, attr, output, value):
        # This is overridden by subclasses
        return self._set_iio_attr(channel_name, attr, output, value)

    def _get_iio_attr_single(self, channel_name, attr, output):
        # This is overridden by subclasses
        return self._get_iio_attr(channel_name, attr, output)

    def _set_iio_attr_single(self, channel_name, attr, output, value):
        # This is overridden by subclasses
        return self._set_iio_attr(channel_name, attr, output, value)

    def _get_iio_dev_attr_single(self, attr):
        # This is overridden by subclasses
        return self._get_iio_dev_attr(attr)

    def _set_iio_dev_attr_single(self, attr, value):
        # This is overridden by subclasses
        return self._set_iio_dev_attr(attr, value)

    def _get_iio_dev_attr_str_single(self, attr):
        # This is overridden by subclasses
        return self._get_iio_dev_attr_str(attr)

    def _set_iio_dev_attr_str_single(self, attr, value):
        # This is overridden by subclasses
        return self._set_iio_dev_attr_str(attr, value)

    @property
    def path_map(self):
        """path_map: Map of channelizers both coarse and fine to
        individual driver channel names
        """
        return self._path_map

    def write_pfilt_config(self, value):
        """Load a new PFILT configuration file
        Input is path to PFILT configuration file. Please see
        driver documentation about PFILT generation and limitations
        """
        with open(value, "r") as file:
            data = file.read()
        self._set_iio_dev_attr_str("filter_fir_config", data)

    # we cannot really get the profile. The driver will just throw EPERM
    pfilt_config = property(None, write_pfilt_config)

    @property
    def rx_channel_nco_frequencies(self):
        """rx_channel_nco_frequencies: Receive path fine DDC NCO frequencies"""
        return self._get_iio_attr_vec(
            self._rx_fine_ddc_channel_names, "channel_nco_frequency", False
        )

    @rx_channel_nco_frequencies.setter
    def rx_channel_nco_frequencies(self, value):
        self._set_iio_attr_int_vec(
            self._rx_fine_ddc_channel_names, "channel_nco_frequency", False, value
        )

    @property
    def rx_channel_nco_phases(self):
        """rx_channel_nco_phases: Receive path fine DDC NCO phases"""
        return self._get_iio_attr_vec(
            self._rx_fine_ddc_channel_names, "channel_nco_phase", False
        )

    @rx_channel_nco_phases.setter
    def rx_channel_nco_phases(self, value):
        self._set_iio_attr_int_vec(
            self._rx_fine_ddc_channel_names, "channel_nco_phase", False, value,
        )

    @property
    def rx_main_nco_frequencies(self):
        """rx_main_nco_frequencies: Receive path coarse DDC NCO frequencies"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_frequency", False
        )

    @rx_main_nco_frequencies.setter
    def rx_main_nco_frequencies(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_frequency", False, value,
        )

    @property
    def rx_main_nco_phases(self):
        """rx_main_nco_phases: Receive path coarse DDC NCO phases"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_phase", False
        )

    @rx_main_nco_phases.setter
    def rx_main_nco_phases(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_phase", False, value,
        )

    @property
    def rx_test_mode(self):
        """rx_test_mode: NCO Test Mode"""
        return self._get_iio_attr_str_single("voltage0_i", "test_mode", False)

    @rx_test_mode.setter
    def rx_test_mode(self, value):
        self._set_iio_attr_single(
            "voltage0_i", "test_mode", False, value,
        )

    @property
    def rx_nyquist_zone(self):
        """rx_nyquist_zone: ADC nyquist zone. Options are: odd, even"""
        return self._get_iio_attr_str_vec(
            self._rx_coarse_ddc_channel_names, "nyquist_zone", False
        )

    @rx_nyquist_zone.setter
    def rx_nyquist_zone(self, value):
        self._set_iio_attr_str_vec(
            self._rx_coarse_ddc_channel_names, "nyquist_zone", False, value,
        )

    @property
    def rx_main_6dB_digital_gains(self):
        """rx_main_6dB_digital_gains: Enable 6dB of gain per CDDC"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_6db_digital_gain_en", False
        )

    @rx_main_6dB_digital_gains.setter
    def rx_main_6dB_digital_gains(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_6db_digital_gain_en", False, value,
        )

    @property
    def rx_channel_6dB_digital_gains(self):
        """rx_channel_6dB_digital_gains: Enable 6dB of gain per FDDC"""
        return self._get_iio_attr_vec(
            self._rx_fine_ddc_channel_names, "channel_6db_digital_gain_en", False
        )

    @rx_channel_6dB_digital_gains.setter
    def rx_channel_6dB_digital_gains(self, value):
        self._set_iio_attr_int_vec(
            self._rx_fine_ddc_channel_names,
            "channel_6db_digital_gain_en",
            False,
            value,
        )

    @property
    def rx_main_nco_ffh_index(self):
        """rx_main_nco_ffh_index: Receive path coarse DDC NCO index in range [0,15]"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_ffh_index", False
        )

    @rx_main_nco_ffh_index.setter
    def rx_main_nco_ffh_index(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_ffh_index", False, value,
        )

    @property
    def rx_main_nco_ffh_select(self):
        """rx_main_nco_ffh_select: Receive path coarse DDC NCO select in range [0,15]"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_ffh_select", False
        )

    @rx_main_nco_ffh_select.setter
    def rx_main_nco_ffh_select(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_nco_ffh_select", False, value,
        )

    @property
    def rx_main_ffh_mode(self):
        """rx_main_ffh_mode: ADC FFH mode. Options are:
        instantaneous_update, synchronous_update_by_transfer_bit,
        synchronous_update_by_gpio
        """
        return self._get_iio_attr_str_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_mode", False
        )

    @rx_main_ffh_mode.setter
    def rx_main_ffh_mode(self, value):
        self._set_iio_attr_str_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_mode", False, value,
        )

    @property
    def rx_main_ffh_trig_hop_en(self):
        """rx_main_ffh_trig_hop_en: Enable triggered hopping for CDDC NCO"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_trig_hop_en", False
        )

    @rx_main_ffh_trig_hop_en.setter
    def rx_main_ffh_trig_hop_en(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_trig_hop_en", False, value,
        )

    @property
    def rx_main_ffh_gpio_mode_enable(self):
        """rx_main_ffh_gpio_mode_enable: Enablles GPIO controlled frequency hopping"""
        return self._get_iio_attr_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_gpio_mode_en", False
        )

    @rx_main_ffh_gpio_mode_enable.setter
    def rx_main_ffh_gpio_mode_enable(self, value):
        self._set_iio_attr_int_vec(
            self._rx_coarse_ddc_channel_names, "main_ffh_gpio_mode_en", False, value,
        )

    @property
    def tx_channel_nco_frequencies(self):
        """tx_channel_nco_frequencies: Transmit path fine DUC NCO frequencies"""
        return self._get_iio_attr_vec(
            self._tx_fine_duc_channel_names, "channel_nco_frequency", True
        )

    @tx_channel_nco_frequencies.setter
    def tx_channel_nco_frequencies(self, value):
        self._set_iio_attr_int_vec(
            self._tx_fine_duc_channel_names, "channel_nco_frequency", True, value
        )

    @property
    def tx_channel_nco_phases(self):
        """tx_channel_nco_phases: Transmit path fine DUC NCO phases"""
        return self._get_iio_attr_vec(
            self._tx_fine_duc_channel_names, "channel_nco_phase", True
        )

    @tx_channel_nco_phases.setter
    def tx_channel_nco_phases(self, value):
        self._set_iio_attr_int_vec(
            self._tx_fine_duc_channel_names, "channel_nco_phase", True, value,
        )

    @property
    def tx_channel_nco_test_tone_en(self):
        """tx_channel_nco_test_tone_en: Transmit path fine DUC NCO test tone enable"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "channel_nco_test_tone_en", True
        )

    @tx_channel_nco_test_tone_en.setter
    def tx_channel_nco_test_tone_en(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "channel_nco_test_tone_en", True, value,
        )

    @property
    def tx_channel_nco_test_tone_scales(self):
        """tx_channel_nco_test_tone_scales: Transmit path fine DUC NCO test tone scale"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "channel_nco_test_tone_scale", True
        )

    @tx_channel_nco_test_tone_scales.setter
    def tx_channel_nco_test_tone_scales(self, value):
        self._set_iio_attr_float_vec(
            self._tx_coarse_duc_channel_names,
            "channel_nco_test_tone_scale",
            True,
            value,
        )

    @property
    def tx_channel_nco_gain_scales(self):
        """tx_channel_nco_gain_scales Transmit path fine DUC NCO gain scale"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "channel_nco_gain_scale", True
        )

    @tx_channel_nco_gain_scales.setter
    def tx_channel_nco_gain_scales(self, value):
        self._set_iio_attr_float_vec(
            self._tx_coarse_duc_channel_names, "channel_nco_gain_scale", True, value,
        )

    @property
    def tx_main_nco_frequencies(self):
        """tx_main_nco_frequencies: Transmit path coarse DUC NCO frequencies"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_frequency", True
        )

    @tx_main_nco_frequencies.setter
    def tx_main_nco_frequencies(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_frequency", True, value,
        )

    @property
    def tx_main_nco_phases(self):
        """tx_main_nco_phases: Transmit path coarse DUC NCO phases"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_phase", True
        )

    @tx_main_nco_phases.setter
    def tx_main_nco_phases(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_phase", True, value,
        )

    @property
    def tx_main_nco_test_tone_en(self):
        """tx_main_nco_test_tone_en: Transmit path coarse DUC NCO test tone enable"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_test_tone_en", True
        )

    @tx_main_nco_test_tone_en.setter
    def tx_main_nco_test_tone_en(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_test_tone_en", True, value,
        )

    @property
    def tx_main_nco_test_tone_scales(self):
        """tx_main_nco_test_tone_scales: Transmit path coarse DUC NCO test tone scale"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_test_tone_scale", True
        )

    @tx_main_nco_test_tone_scales.setter
    def tx_main_nco_test_tone_scales(self, value):
        self._set_iio_attr_float_vec(
            self._tx_coarse_duc_channel_names, "main_nco_test_tone_scale", True, value,
        )

    @property
    def tx_main_ffh_frequency(self):
        """tx_main_ffh_frequency: Transmitter fast frequency hop frequency. This will set
        The NCO frequency of the NCO selected from the bank defined by tx_main_ffh_index
        """
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_frequency", True
        )

    @tx_main_ffh_frequency.setter
    def tx_main_ffh_frequency(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_frequency", True, value,
        )

    @property
    def tx_main_ffh_index(self):
        """tx_main_ffh_index: Transmitter fast frequency hop NCO bank index  in range [0,30]"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_index", True
        )

    @tx_main_ffh_index.setter
    def tx_main_ffh_index(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_index", True, value,
        )

    @property
    def tx_main_nco_ffh_select(self):
        """tx_main_nco_ffh_select: Transmit path coarse DDC NCO select in range [0,30]"""
        return self._get_iio_attr_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_select", True
        )

    @tx_main_nco_ffh_select.setter
    def tx_main_nco_ffh_select(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "main_nco_ffh_select", True, value,
        )

    @property
    def tx_main_ffh_mode(self):
        """tx_main_ffh_mode: Set hop transition mode of NCOs Options are:
        phase_continuous, phase_incontinuous, and phase_coherent
        """
        return self._get_iio_attr_str_vec(
            self._tx_coarse_duc_channel_names, "main_ffh_mode", True
        )

    @tx_main_ffh_mode.setter
    def tx_main_ffh_mode(self, value):
        self._set_iio_attr_str_vec(
            self._tx_coarse_duc_channel_names, "main_ffh_mode", True, value,
        )

    @property
    def tx_main_ffh_gpio_mode_enable(self):
        """tx_main_ffh_gpio_mode_enable: Enablles GPIO controlled frequency hopping"""
        return self._get_iio_attr_single("voltage0_i", "main_ffh_gpio_mode_en", True)

    @tx_main_ffh_gpio_mode_enable.setter
    def tx_main_ffh_gpio_mode_enable(self, value):
        self._set_iio_attr_single(
            "voltage0_i", "main_ffh_gpio_mode_en", True, value,
        )

    @property
    def tx_dac_en(self):
        """tx_dac_en: Enable DACs"""
        return self._get_iio_attr_vec(self._tx_coarse_duc_channel_names, "en", True)

    @tx_dac_en.setter
    def tx_dac_en(self, value):
        self._set_iio_attr_int_vec(
            self._tx_coarse_duc_channel_names, "en", True, value,
        )

    def set_tx_dac_full_scale_current(self, value):
        """tx_dac_full_scale_current: Set full scale current of DACs. This value
        is in microamps.
        """
        self._set_iio_debug_attr_str(
            "dac-full-scale-current-ua", str(value), self._rxadc
        )

    # we cannot read current as the driver will just throw EPERM
    tx_dac_full_scale_current = property(None, set_tx_dac_full_scale_current)

    @property
    def powerdown(self):
        """powerdown: Powerdown and reset the chip

        Support for dynamic powerdown. Writing device attribute
        powerdown with 'Yy1Nn0', or [oO][NnFf] for "on" and "off", will either
        stop the jesd204 fsm, reset the device and power down an optional
        regulator (vdd), or do the opposite in reverse order.
        """
        return self._get_iio_dev_attr_single("powerdown")

    @powerdown.setter
    def powerdown(self, value):
        self._set_iio_dev_attr_single(
            "powerdown", value,
        )

    @property
    def loopback_mode(self):
        """loopback_mode: Enable loopback mode RX->TX

        When enabled JESD RX FIFO is connected to JESD TX FIFO,
        making the entire datasource for the TX path the RX path. No
        data is passed into the TX path from off-chip when 1. For
        this mode to function correctly the JESD configuration
        between RX and TX must be identical and only use a single
        link.
        """
        return self._get_iio_dev_attr_single("loopback_mode")

    @loopback_mode.setter
    def loopback_mode(self, value):
        self._set_iio_dev_attr_single(
            "loopback_mode", value,
        )

    @property
    def tx_ddr_offload(self):
        """tx_ddr_offload: Enable DDR offload

        When true the DMA will pass buffers into the BRAM FIFO for data repeating.
        This is necessary when operating at high DAC sample rates. This can reduce
        the maximum buffer size but data passed to DACs in cyclic mode will not
        underflow due to memory bottlenecks.
        """
        return self._get_iio_debug_attr("pl_ddr_fifo_enable", self._txdac)

    @tx_ddr_offload.setter
    def tx_ddr_offload(self, value):
        self._set_iio_debug_attr_str("pl_ddr_fifo_enable", str(value * 1), self._txdac)

    @property
    def rx_sample_rate(self):
        """rx_sampling_frequency: Sample rate after decimation"""
        return self._get_iio_attr_single("voltage0_i", "sampling_frequency", False)

    @property
    def adc_frequency(self):
        """adc_frequency: ADC frequency in Hz"""
        return self._get_iio_attr_single("voltage0_i", "adc_frequency", False)

    @property
    def tx_sample_rate(self):
        """tx_sampling_frequency: Sample rate before interpolation"""
        return self._get_iio_attr_single("voltage0_i", "sampling_frequency", True)

    @property
    def dac_frequency(self):
        """dac_frequency: DAC frequency in Hz"""
        return self._get_iio_attr_single("voltage0_i", "dac_frequency", True)

    @property
    def jesd204_fsm_ctrl(self):
        """jesd204_fsm_ctrl: jesd204-fsm control"""
        return self._get_iio_dev_attr("jesd204_fsm_ctrl", self._rxadc)

    @jesd204_fsm_ctrl.setter
    def jesd204_fsm_ctrl(self, value):
        self._set_iio_dev_attr("jesd204_fsm_ctrl", value, self._rxadc)

    @property
    def jesd204_fsm_resume(self):
        """jesd204_fsm_resume: jesd204-fsm resume"""
        return self._get_iio_dev_attr("jesd204_fsm_resume", self._rxadc)

    @jesd204_fsm_resume.setter
    def jesd204_fsm_resume(self, value):
        self._set_iio_dev_attr_str("jesd204_fsm_resume", value, self._rxadc)

    @property
    def jesd204_fsm_state(self):
        """jesd204_fsm_state: jesd204-fsm state"""
        return self._get_iio_dev_attr_str("jesd204_fsm_state", self._rxadc)

    @property
    def jesd204_fsm_paused(self):
        """jesd204_fsm_paused: jesd204-fsm paused"""
        return self._get_iio_dev_attr("jesd204_fsm_paused", self._rxadc)

    @property
    def jesd204_fsm_error(self):
        """jesd204_fsm_error: jesd204-fsm error"""
        return self._get_iio_dev_attr("jesd204_fsm_error", self._rxadc)

    @property
    def jesd204_device_status(self):
        """jesd204_device_status: Device jesd204 link status information"""
        return self._get_iio_debug_attr_str("status", self._rxadc)

    @property
    def jesd204_device_status_check(self):
        """jesd204_device_status_check: Device jesd204 link status check

        Returns 'True' in case error conditions are detected, 'False' otherwise
        """
        stat = self._get_iio_debug_attr_str("status", self._rxadc)

        for s in stat.splitlines(0):
            if "JRX" in s:
                if "204C" in s:
                    if "Link is good" not in s:
                        return True
                elif "204B" in s:
                    if "0x0 lanes in DATA" in s:
                        return True
            elif "JTX" in s:
                if any(
                    substr in s
                    for substr in [" asserted", "unlocked", "lost", "invalid"]
                ):
                    return True
        return False

    @property
    def chip_version(self):
        """chip_version: Chip version information"""
        return self._get_iio_debug_attr_str("chip_version", self._rxadc)

    @property
    def api_version(self):
        """api_version: API version"""
        return self._get_iio_debug_attr_str("api_version", self._rxadc)



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.context_manager import context_manager
from adi.rx_tx import tx


class ad5627(tx, context_manager):
    """ AD5627 Low Power Dual nanoDAC """

    _complex_data = False
    _tx_channel_names = ["voltage0", "voltage1"]
    _device_name = ""

    def __init__(self, uri=""):
        context_manager.__init__(self, uri, self._device_name)
        self._txdac = self._ctx.find_device("ad5627")
        tx.__init__(self)

    @staticmethod
    def _adp_bias_volts_to_raw_convert(value, inverse):
        """Convert ADP bias from user values to internal values (or reverse)."""
        ret = 0.0
        if inverse:
            ret = -((value * 5 * 18.18) / 4096) + 122
        else:
            ret = ((-122 - value) * 4096) / (5 * 18.18)
        return ret

    @property
    def apdbias(self):
        """Get the APD Bias."""
        bias = self._get_iio_attr("voltage0", "raw", True, self._txdac)
        return self._adp_bias_volts_to_raw_convert(bias, True)

    @apdbias.setter
    def apdbias(self, value):
        """Set the APD Bias."""
        bias = self._adp_bias_volts_to_raw_convert(value, False)
        self._set_iio_attr_float("voltage0", "raw", True, bias, self._txdac)

    @staticmethod
    def _tilt_volts_to_raw_convert(value, inverse):
        """Convert tilt voltage from user values to internal values (or reverse)."""
        ret = 0.0
        if inverse:
            ret = (value * 5) / 4096
        else:
            ret = (value * 4096) / 5
        return ret

    @property
    def tiltvoltage(self):
        """Get the Tilt Voltage."""
        voltage = self._get_iio_attr("voltage1", "raw", True, self._txdac)
        return self._tilt_volts_to_raw_convert(voltage, True)

    @tiltvoltage.setter
    def tiltvoltage(self, voltage):
        """Set the Tilt Voltage."""
        voltage = self._tilt_volts_to_raw_convert(voltage, False)
        self._set_iio_attr_float("voltage1", "raw", True, voltage, self._txdac)



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import List

from adi.attribute import attribute
from adi.context_manager import context_manager


class tdd(context_manager, attribute):

    """TDD Controller"""

    _device_name: str = ""

    def __init__(self, uri=""):
        """TDD Controller"""
        context_manager.__init__(self, uri, self._device_name)
        self._ctrl = self._ctx.find_device("axi-core-tdd")

    @property
    def frame_length_ms(self) -> float:
        """frame_length_ms: TDD frame length in ms"""
        return self._get_iio_dev_attr("frame_length_ms")

    @frame_length_ms.setter
    def frame_length_ms(self, value: float):
        self._set_iio_dev_attr("frame_length_ms", value)

    @property
    def frame_length_raw(self) -> float:
        """frame_length_raw: TDD frame length in cycles of the input clock"""
        return self._get_iio_dev_attr("frame_length_raw")

    @frame_length_raw.setter
    def frame_length_raw(self, value: float):
        self._set_iio_dev_attr("frame_length_raw", value)

    @property
    def burst_count(self) -> int:
        """burst_count: Amount of frames to produce.
        Should be 0 <= burst_count <= 255, where 0 means don't stop
        """
        return self._get_iio_dev_attr("burst_count")

    @burst_count.setter
    def burst_count(self, value: int):
        self._set_iio_dev_attr("burst_count", value)

    @property
    def counter_int(self) -> int:
        """counter_init: Internal counter start value"""
        return self._get_iio_dev_attr("counter_int")

    @counter_int.setter
    def counter_int(self, value: int):
        self._set_iio_dev_attr("counter_int", value)

    @property
    def dma_gateing_mode(self) -> str:
        """dma_gateing_mode: Which of the two DMA (dp) ports should be gated"""
        return self._get_iio_dev_attr("dma_gateing_mode")

    @dma_gateing_mode.setter
    def dma_gateing_mode(self, value: str):
        return self._set_iio_dev_attr("dma_gateing_mode", value)

    @property
    def en(self) -> bool:
        """en: Enable or disable the TDD engine"""
        return bool(int(self._get_iio_dev_attr("en")))

    @en.setter
    def en(self, value: bool):
        self._set_iio_dev_attr("en", int(value))

    @property
    def en_mode(self) -> str:
        """en_mode: In which mode the TDD engine should run"""
        return self._get_iio_dev_attr("en_mode")

    @en_mode.setter
    def en_mode(self, value: str):
        return self._set_iio_dev_attr("en_mode", value)

    @property
    def secondary(self) -> bool:
        """secondary: Enable secondary times. This allows one signal to go high
        twice at two times within a single frame
        """
        return bool(int(self._get_iio_dev_attr("secondary")))

    @secondary.setter
    def secondary(self, value: bool):
        self._set_iio_dev_attr("secondary", int(value))

    #
    # TDD channels
    #

    def __get_tdd_channel(self, name: str, output: bool, raw: bool) -> List:
        return [
            self._get_iio_attr(
                "data{}".format(c),
                "{}{}_{}".format(name, d, "raw" if raw else "ms"),
                output,
            )
            for d, c in [("on", 0), ("off", 0), ("on", 1), ("off", 1)]
        ]

    def __set_tdd_channel(
        self, name: str, output: bool, raw: bool, values: List
    ) -> List:
        if len(values) != 4:
            raise RuntimeError("Expected four values, received {}".format(len(values)))

        for v, d, c in zip(values, ["on", "off", "on", "off"], [0, 0, 1, 1]):
            self._set_iio_attr(
                "data{}".format(c),
                "{}{}_{}".format(name, d, "raw" if raw else "ms"),
                output,
                v,
            )

    @property
    def tx_dma_raw(self) -> List[int]:
        """tx_dma_raw: TX DMA port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("dp_", True, True)

    @tx_dma_raw.setter
    def tx_dma_raw(self, values: List[int]):
        self.__set_tdd_channel("dp_", True, True, values)

    @property
    def rx_dma_raw(self) -> List[int]:
        """rx_dma_raw: RX DMA port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("dp_", False, True)

    @rx_dma_raw.setter
    def rx_dma_raw(self, values: List[int]):
        self.__set_tdd_channel("dp_", False, True, values)

    @property
    def tx_dma_ms(self) -> List[int]:
        """tx_dma_ms: TX DMA port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("dp_", True, False)

    @tx_dma_ms.setter
    def tx_dma_ms(self, values: List[int]):
        self.__set_tdd_channel("dp_", True, False, values)

    @property
    def rx_dma_ms(self) -> List[int]:
        """rx_dma_ms: RX DMA port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("dp_", False, False)

    @rx_dma_ms.setter
    def rx_dma_ms(self, values: List[int]):
        self.__set_tdd_channel("dp_", False, False, values)

    @property
    def tx_rf_raw(self) -> List[int]:
        """tx_rf_raw: TX RF port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("", True, True)

    @tx_rf_raw.setter
    def tx_rf_raw(self, values: List[int]):
        self.__set_tdd_channel("", True, True, values)

    @property
    def rx_rf_raw(self) -> List[int]:
        """rx_rf_raw: RX RF port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("", False, True)

    @rx_rf_raw.setter
    def rx_rf_raw(self, values: List[int]):
        self.__set_tdd_channel("", False, True, values)

    @property
    def tx_rf_ms(self) -> List[int]:
        """tx_rf_ms: TX RF port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("", True, False)

    @tx_rf_ms.setter
    def tx_rf_ms(self, values: List[int]):
        self.__set_tdd_channel("", True, False, values)

    @property
    def rx_rf_ms(self) -> List[int]:
        """rx_rf_ms: RX RF port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("", False, False)

    @rx_rf_ms.setter
    def rx_rf_ms(self, values: List[int]):
        self.__set_tdd_channel("", False, False, values)

    @property
    def tx_vco_raw(self) -> List[int]:
        """tx_vco_raw: TX VCO port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("vco_", True, True)

    @tx_vco_raw.setter
    def tx_vco_raw(self, values: List[int]):
        self.__set_tdd_channel("vco_", True, True, values)

    @property
    def rx_vco_raw(self) -> List[int]:
        """rx_vco_raw: RX VCO port timing parameters in clock cycles.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("vco_", False, True)

    @rx_vco_raw.setter
    def rx_vco_raw(self, values: List[int]):
        self.__set_tdd_channel("vco_", False, True, values)

    @property
    def tx_vco_ms(self) -> List[int]:
        """tx_vco_ms: TX VCO port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("vco_", True, False)

    @tx_vco_ms.setter
    def tx_vco_ms(self, values: List[int]):
        self.__set_tdd_channel("vco_", True, False, values)

    @property
    def rx_vco_ms(self) -> List[int]:
        """rx_vco_ms: RX VCO port timing parameters in ms.
        List of four values: [primary_on, primary_off, secondary_on, secondary_off]
        """
        return self.__get_tdd_channel("vco_", False, False)

    @rx_vco_ms.setter
    def rx_vco_ms(self, values: List[int]):
        self.__set_tdd_channel("vco_", False, False, values)



# Copyright (C) 2020 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .sshfs import sshfs


class jesd:
    """JESD Monitoring"""

    def __init__(self, address, username="root", password="analog"):
        if "ip:" in address:
            address = address[3:]
        self.address = address
        self.username = username
        self.password = password
        self.rootdir = "/sys/bus/platform/devices/"

        # Connect
        self.fs = sshfs(address, username, password)

        self.find_jesd_dir()
        self.find_lanes()

    def find_lanes(self):
        self.lanes = {}
        for dr in self.dirs:
            if "-rx" in dr:
                self.lanes[dr] = []
                lanIndx = 0
                while 1:
                    li = "/lane{}_info".format(lanIndx)
                    if self.fs.isfile(self.rootdir + dr + li):
                        self.lanes[dr].append(li)
                        lanIndx += 1
                    else:
                        break

    def find_jesd_dir(self):
        dirs = self.fs.listdir(self.rootdir)
        self.dirs = []
        for dr in dirs:
            if "jesd" in dr:
                self.dirs.append(dr)

    def decode_status(self, status):
        status = status.replace(",", "\n")
        status = status.split("\n")
        link_status = {}
        for s in status:
            if ":" in s:
                o = s.split(":")
                link_status[o[0].strip().replace("/", "")] = o[1].strip()
            if "Link is" in s:
                link_status["enabled"] = s.split(" ")[-1].strip()

        return link_status

    def get_status(self, dr):
        return self.fs.gettext(self.rootdir + dr + "/status")

    def get_dev_lane_info(self, dr):
        return {
            ldir.replace("/", ""): self.decode_status(
                self.fs.gettext(self.rootdir + dr + "/" + ldir)
            )
            for ldir in self.lanes[dr]
        }

    def get_all_link_statuses(self):
        statuses = dict()
        for dr in self.dirs:
            if "-rx" in dr:
                statuses[dr] = self.get_dev_lane_info(dr)
        return statuses

    def get_all_statuses(self):
        return {dr: self.decode_status(self.get_status(dr)) for dr in self.dirs}



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import numpy as np
from adi.ad9166 import ad9166 as ad9166_adi  # pyadi-iio library
from adi.context_manager import context_manager

try:
    import ad9166 as libad9166  # helper library for calibration function
except ImportError:
    libad9166 = None


class cn0511(ad9166_adi):
    """ CN0511 Raspberry Pi Hat Signal Generator """

    def __init__(self, uri=""):
        context_manager.__init__(self, uri, self._device_name)
        ad9166_adi.__init__(self, uri=uri)
        self._trim_dac_ch = "voltage0"
        self._trim_dac = self._ctx.find_device("ad5693r")
        self._amp = self._ctx.find_device("ad9166-amp")
        self.__calibrated_dev = self._ctx.find_device("ad9166")
        self.__ch = self.__calibrated_dev.find_channel("altvoltage0", True)
        if libad9166:
            self.__calibration_data = libad9166.find_calibration_data(
                self._ctx, "cn0511"
            )

        self.FIR85_enable = True
        self.sample_rate = 6000000000

    @property
    def trim_frequency_raw(self):
        """ trim_frequency_raw: modify output frequency of cn0511 in small steps"""
        return self._get_iio_attr(self._trim_dac_ch, "raw", True, self._trim_dac)

    @trim_frequency_raw.setter
    def trim_frequency_raw(self, value):
        self._set_iio_attr(self._trim_dac_ch, "raw", True, value, self._trim_dac)

    @property
    def amp_enable(self):
        """ amp_enable: Enable or Disable the CN0511 ad9166 amplifier """
        return True if (self._get_iio_dev_attr("en", _ctrl=self._amp) == 1) else False

    @amp_enable.setter
    def amp_enable(self, value=True):
        if value:
            val = 1
        else:
            val = 0
        self._set_iio_dev_attr_str("en", val, _ctrl=self._amp)

    @property
    def amplitude_cal(self):
        """amplitude_cal: CN0511 amplitude calibration

           Options:
               True: If you set this to true, the output is calibrated.
               False: Nothing happens.
        """
        return None

    @amplitude_cal.setter
    def amplitude_cal(self, value=True):
        if libad9166:
            if value:
                libad9166.device_set_iofs(
                    self.__calibrated_dev, self.__calibration_data, self.frequency
                )
        else:
            print("Warning: Missing libad9166, calibration failed.")

    @property
    def calibrated_output(self):
        """calibrated_output: ["desired_output_amplitude_in_dbm", "desired_output_frequency_in_Hz"]]"""
        if libad9166:
            return [int(20 * np.log10(self.raw / (2 ** 15))), self.frequency]
        else:
            print("Warning: Missing libad9166, calibration failed.")
            return [None, None]

    @calibrated_output.setter
    def calibrated_output(self, value):
        if libad9166:
            libad9166.set_amplitude(self.__calibrated_dev, value[0])
            libad9166.set_frequency(self.__ch, value[1])
            libad9166.device_set_iofs(
                self.__calibrated_dev, self.__calibration_data, value[1]
            )
        else:
            print("Warning: Missing libad9166, calibration failed.")



# Copyright (C) 2019 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
from adi.context_manager import context_manager


class ad9166(attribute, context_manager):
    """ AD9166 Vector Signal Generator """

    _complex_data = False
    channel = []  # type: ignore
    _device_name = ""
    _temperatureRef = 32.0
    _temperatureRefCode = 0x8008
    _temperatureM = (_temperatureRef + 190) / (_temperatureRefCode / 1000)

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)
        self._ctrl = self._ctx.find_device("ad9166")
        self._txdac = self._ctx.find_device("ad9166")
        self._temp_sensor_name = "temp0"
        self._dac0_name = "altvoltage0"

    @property
    def sample_rate(self):
        """ sample_rate: Sets sampling frequency of the AD916x """
        return self._get_iio_dev_attr("sampling_frequency")

    @sample_rate.setter
    def sample_rate(self, value):
        self._set_iio_dev_attr_str("sampling_frequency", value)

    @property
    def sample_rate_available(self):
        return [5000000000, 5898240000, 6000000000]

    @property
    def temperature(self):
        """ temperature: Returns the AD916x Chip Temperature in Celsius """
        temp = 0
        if self.temperature_enable:
            try:
                temp = (
                    self._get_iio_attr(self._temp_sensor_name, "input", False) / 1000.0
                )
            except Exception:
                # Use temporary single point calibration if not calibrated
                self.temperature_cal = 34.0
                temp = (
                    self._get_iio_attr(self._temp_sensor_name, "input", False) / 1000.0
                )
            return temp
        return None

    @property
    def temperature_code(self):
        """ temperature_code: Returns the AD916x Chip Temperature ADC code """
        return self._get_iio_attr(self._temp_sensor_name, "raw", False)

    @property
    def temperature_enable(self):
        """ temperature_enable: AD9166 Chip Temperature Measurement Enable

            Options:
                True: Temperature measurement is enabled
                False: Temperature measurement is disabled
        """
        reg = self._ctrl.reg_read(0x135)
        return ((reg and 0x01)) == 0x01

    @temperature_enable.setter
    def temperature_enable(self, value=True):
        if value:
            self._ctrl.reg_write(0x135, 0xA1)
        else:
            self._ctrl.reg_write(0x135, 0xA0)

    @property
    def temperature_cal(self):
        """ temperature_cal: AD9166 Chip Temperature single point calibration value.
        Enter the ambient temperature in degree Celsius.
        """
        return None

    @temperature_cal.setter
    def temperature_cal(self, value=32.0):
        try:
            val = int(round(value * 1000))
            self._set_iio_attr(self._temp_sensor_name, "single_point_calib", False, val)
        except Exception as ex:
            raise ex

    @property
    def nco_enable(self):
        """ nco_enable: AD9166 NCO Modulation Enable:

            Options:
                True: NCO Modulation is enabled
                False: NCO Modulation is disabled
        """
        tmp_reg = self._ctrl.reg_read(0x111) & 0x40
        return tmp_reg != 0

    @nco_enable.setter
    def nco_enable(self, value=True):
        tmp_reg = self._ctrl.reg_read(0x111) & 0xBF
        if value:
            # DATAPATH_CFG (0x111) Bit6=1
            self._ctrl.reg_write(0x111, tmp_reg | 0x40)

        else:
            # DATAPATH_CFG (0x111) Bit6=0
            self._ctrl.reg_write(0x111, tmp_reg)

    @property
    def FIR85_enable(self):
        """ FIR85_enable: AD9166 FIR85 Filter Enable:

            Options:
                True: FIR85 Filter is enabled
                False: FIR85 Filter is disabled
        """
        tmp_reg = self._ctrl.reg_read(0x111) & 0x01
        return tmp_reg != 0

    @FIR85_enable.setter
    def FIR85_enable(self, value=True):
        tmp_reg = self._ctrl.reg_read(0x111) & 0xFE
        if value:
            # DATAPATH_CFG (0x111) Bit0=1
            self._ctrl.reg_write(0x111, tmp_reg | 0x01)
        else:
            # DATAPATH_CFG (0x111) Bit0=0
            self._ctrl.reg_write(0x111, tmp_reg)

    @property
    def tx_enable(self):
        """ tx_enable: AD9166 TX Enable

            Options:
                True: TX is enabled (Datapath is connected to DAC)
                False: TX is disabled or  (DAC input is zeroed)
        """
        tmp_reg = self._ctrl.reg_read(0x03F) & 0x80
        return tmp_reg != 0

    @tx_enable.setter
    def tx_enable(self, value=True):
        # Supports only SPI Data Post Enable/disable.
        # TODO: Implement other tx enable ways.

        tmp_reg = self._ctrl.reg_read(0x03F) & 0x7F
        if value:
            # TX_ENABLE (0x3F) Bit7=1
            self._ctrl.reg_write(0x03F, tmp_reg | 0x80)
        else:
            # TX_ENABLE (0x3F) Bit7=0
            self._ctrl.reg_write(0x03F, tmp_reg)

    @property
    def frequency(self):
        """ frequency: AD916x channel nco frequency value in hz."""
        return self._get_iio_attr(self._dac0_name, "nco_frequency", True)

    @frequency.setter
    def frequency(self, value):
        return self._set_iio_attr(self._dac0_name, "nco_frequency", True, value)

    @property
    def raw(self):
        """ raw: AD916x channel raw value. Integer range 0-32767. """
        return self._get_iio_attr(self._dac0_name, "raw", True)

    @raw.setter
    def raw(self, value):
        return self._set_iio_attr(self._dac0_name, "raw", True, value)



# Copyright (C) 2021 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from decimal import Decimal

import numpy as np
from adi.attribute import attribute
from adi.context_manager import context_manager
from adi.rx_tx import rx


class adxl355(rx, context_manager, attribute):
    """ ADXL355 3-axis accelerometer """

    _device_name = "adxl355"
    _rx_unbuffered_data = True
    _rx_data_si_type = float

    def __init__(self, uri=""):

        context_manager.__init__(self, uri, self._device_name)
        self._ctrl = self._ctx.find_device("adxl355")
        self.accel_x = self._channel(self._ctrl, "accel_x")
        self.accel_y = self._channel(self._ctrl, "accel_y")
        self.accel_z = self._channel(self._ctrl, "accel_z")
        self.temp = self._tempchannel(self._ctrl, "temp")
        self._rxadc = self._ctx.find_device("adxl355")
        self._rx_channel_names = ["accel_x", "accel_y", "accel_z"]
        rx.__init__(self)

    @property
    def current_timestamp_clock(self):
        """Current timestamp clock"""
        return self._get_iio_dev_attr("current_timestamp_clock")

    def to_degrees(self, raw):
        """Convert raw to degrees Celsius"""
        return (raw + self.temp.offset) * self.temp.scale / 1000.0

    class _tempchannel(attribute):
        """ADXL355 temperature channel"""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def offset(self):
            """ADXL355 temperature offset value"""
            return self._get_iio_attr(self.name, "offset", False)

        @property
        def raw(self):
            """ADXL355 temperature raw value"""
            return self._get_iio_attr(self.name, "raw", False)

        @property
        def scale(self):
            """ADXL355 channel scale value"""
            return self._get_iio_attr(self.name, "scale", False)

    class _channel(attribute):
        """ADXL355 acceleration channel"""

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def calibbias(self):
            """ADXL355 channel offset"""
            return self._get_iio_attr(self.name, "calibbias", False)

        @calibbias.setter
        def calibbias(self, value):
            self._set_iio_attr(self.name, "calibbias", False, value)

        @property
        def filter_high_pass_3db_frequency(self):
            """ADXL355 highpass filter cutoff frequency"""
            return self._get_iio_attr(
                self.name, "filter_high_pass_3db_frequency", False
            )

        @filter_high_pass_3db_frequency.setter
        def filter_high_pass_3db_frequency(self, value):
            self._set_iio_attr(
                self.name,
                "filter_high_pass_3db_frequency",
                False,
                str(Decimal(value).real),
            )

        @property
        def filter_high_pass_3db_frequency_available(self):
            """Provides all available highpass filter cutoff frequency settings for the ADXL355 channels"""
            return self._get_iio_attr(
                self.name, "filter_high_pass_3db_frequency_available", False
            )

        @property
        def raw(self):
            """ADXL355 channel raw value"""
            return self._get_iio_attr(self.name, "raw", False)

        @property
        def scale(self):
            """ADXL355 channel scale(gain)"""
            return float(self._get_iio_attr_str(self.name, "scale", False))

        @property
        def sampling_frequency_available(self):
            """Provides all available sampling frequency settings for the ADXL355 channels"""
            return self._get_iio_attr(self.name, "sampling_frequency_available", False)

        @property
        def sampling_frequency(self):
            """ADXL355 per-channel sampling frequency"""
            return float(self._get_iio_attr_str(self.name, "sampling_frequency", False))

        @sampling_frequency.setter
        def sampling_frequency(self, value):
            self._set_iio_attr(
                self.name, "sampling_frequency", False, str(Decimal(value).real)
            )



# Copyright (C) 2023 Analog Devices, Inc.
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#     - Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     - Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the
#       distribution.
#     - Neither the name of Analog Devices, Inc. nor the names of its
#       contributors may be used to endorse or promote products derived
#       from this software without specific prior written permission.
#     - The use of this software may or may not infringe the patent rights
#       of one or more patent holders.  This license does not release you
#       from the requirement that you obtain separate licenses from these
#       patent holders to use this software.
#     - Use of the software either in source or binary form, must be run
#       on or directly connected to an Analog Devices Inc. component.
#
# THIS SOFTWARE IS PROVIDED BY ANALOG DEVICES "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A
# PARTICULAR PURPOSE ARE DISCLAIMED.
#
# IN NO EVENT SHALL ANALOG DEVICES BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, INTELLECTUAL PROPERTY
# RIGHTS, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
# THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from adi.attribute import attribute
from adi.context_manager import context_manager


class ltc2688(context_manager, attribute):
    """ LTC2688 DAC """

    _complex_data = False
    _device_name = "LTC2688"
    vref = 4.096
    channel_names = []

    def __init__(self, uri="ip:analog.local", device_index=0):
        context_manager.__init__(self, uri, self._device_name)

        self._ctrl = self._ctx.find_device("ltc2688")

        for ch in self._ctrl.channels:
            name = ch.id
            self.channel_names.append(name)
            if "toggle_en" in ch.attrs:
                if "symbol" in ch.attrs:
                    setattr(self, name, self._channel_sw_toggle(self._ctrl, name))
                else:
                    setattr(self, name, self._channel_toggle(self._ctrl, name))
            elif "dither_en" in ch.attrs:
                setattr(self, name, self._channel_dither(self._ctrl, name))
            else:
                setattr(self, name, self._channel_standard(self._ctrl, name))

    class _channel_base(attribute):
        """ LTC2688 base channel class """

        def __init__(self, ctrl, channel_name):
            self.name = channel_name
            self._ctrl = ctrl

        @property
        def scale(self):
            """ LTC2688 channel gain """
            return self._get_iio_attr(self.name, "scale", True, self._ctrl)

        @property
        def offset(self):
            """ LTC2688 channel offset """
            return self._get_iio_attr(self.name, "offset", True, self._ctrl)

        @property
        def calibbias(self):
            """ LTC2688 channel calibbias property """
            return self._get_iio_attr(self.name, "calibbias", True, self._ctrl)

        @property
        def calibscale(self):
            """ LTC2688 channel calibbias scale """
            return self._get_iio_attr(self.name, "calibscale", True, self._ctrl)

        @property
        def powerdown(self):
            """ LTC2688 channel powerdown """
            return self._get_iio_attr_str(self.name, "powerdown", True)

        @powerdown.setter
        def powerdown(self, val):
            """ LTC2688 channel powerdown """
            self._set_iio_attr_str(self.name, "powerdown", True, val)

    class _channel_standard(_channel_base):
        def __init__(self, ctrl, channel_name):
            super().__init__(ctrl, channel_name)

        @property
        def raw(self):
            """ LTC2688 channel raw value property """
            return self._get_iio_attr(self.name, "raw", True, self._ctrl)

        @raw.setter
        def raw(self, val):
            """ LTC2688 channel raw value setter """
            raw_span = self._get_iio_attr(self.name, "raw_available", True, self._ctrl)
            if val >= raw_span[0] and val <= raw_span[2] and val % raw_span[1] == 0:
                self._set_iio_attr(self.name, "raw", True, str(int(val)))

        @property
        def volt(self):
            """ LTC2688 channel volt property """
            return ((self.raw + self.offset) * self.scale) * (ltc2688.vref / 4.096)

        @volt.setter
        def volt(self, val):
            """ LTC2688 channel volt setter """
            self.raw = int(((val / self.scale) - self.offset) * (ltc2688.vref / 4.096))

    class _channel_dither(_channel_standard):
        def __init__(self, ctrl, channel_name):
            super().__init__(ctrl, channel_name)

        @property
        def dither_en(self):
            """ LTC2688 channel dither enable flag """
            return self._get_iio_attr(self.name, "dither_en", True, self._ctrl)

        @dither_en.setter
        def dither_en(self, val):
            """ LTC2688 channel dither enable flag setter """
            self._set_iio_attr(self.name, "dither_en", True, val)

        @property
        def dither_frequency(self):
            """ LTC2688 channel dither frequency """
            return self._get_iio_attr(self.name, "dither_frequency", True, self._ctrl)

        @dither_frequency.setter
        def dither_frequency(self, val):
            """ LTC2688 channel dither frequency setter """
            dither_frequency_span = self._get_iio_attr(
                self.name, "dither_frequency_available", True, self._ctrl
            )
            for freq in dither_frequency_span:
                if val == freq:
                    self._set_iio_attr(self.name, "dither_frequency", True, val)
                    break

        @property
        def dither_phase(self):
            """ LTC2688 channel dither phase """
            return self._get_iio_attr(self.name, "dither_phase", True, self._ctrl)

        @dither_phase.setter
        def dither_phase(self, val):
            """ LTC2688 channel dither phase setter """
            dither_phase_span = self._get_iio_attr(
                self.name, "dither_phase_available", True, self._ctrl
            )
            for phase in dither_phase_span:
                if val == phase:
                    self._set_iio_attr(self.name, "dither_phase", True, val)
                    break

        @property
        def dither_raw(self):
            """ LTC2688 channel dither raw value property """
            return self._get_iio_attr(self.name, "dither_raw", True, self._ctrl)

        @dither_raw.setter
        def dither_raw(self, val):
            """ LTC2688 channel dither raw value property setter """
            dither_raw_span = self._get_iio_attr(
                self.name, "dither_raw_available", True, self._ctrl
            )
            if (
                val >= dither_raw_span[0]
                and val <= dither_raw_span[2]
                and val % dither_raw_span[1] == 0
            ):
                self._set_iio_attr(self.name, "dither_raw", True, str(int(val)))

        @property
        def dither_offset(self):
            """ LTC2688 channel dither offset """
            return self._get_iio_attr(self.name, "dither_offset", True, self._ctrl)

        @dither_offset.setter
        def dither_offset(self, val):
            """ LTC2688 channel dither offset setter """
            self._set_iio_attr(self.name, "dither_offset", True, str(int(val)))

    class _channel_toggle(_channel_base):
        def __init__(self, ctrl, channel_name):
            super().__init__(ctrl, channel_name)

        @property
        def toggle_en(self):
            """ LTC2688 channel toggle enable flag """
            return self._get_iio_attr(self.name, "toggle_en", True, self._ctrl)

        @toggle_en.setter
        def toggle_en(self, val):
            """ LTC2688 channel toggle enable flag setter """
            self._set_iio_attr(self.name, "toggle_en", True, val)

        @property
        def raw0(self):
            """ LTC2688 channel toggle state 0 raw value """
            return self._get_iio_attr(self.name, "raw0", True, self._ctrl)

        @raw0.setter
        def raw0(self, val):
            """ LTC2688 channel toggle state 0 raw value setter """
            raw_span = self._get_iio_attr(self.name, "raw_available", True, self._ctrl)
            if val >= raw_span[0] and val <= raw_span[2] and val % raw_span[1] == 0:
                self._set_iio_attr(self.name, "raw0", True, str(int(val)))

        @property
        def raw1(self):
            """ LTC2688 channel toggle state 1 raw value """
            return self._get_iio_attr(self.name, "raw1", True, self._ctrl)

        @raw1.setter
        def raw1(self, val):
            """ LTC2688 channel toggle state 1 raw value setter """
            raw_span = self._get_iio_attr(self.name, "raw_available", True, self._ctrl)
            if val >= raw_span[0] and val <= raw_span[2] and val % raw_span[1] == 0:
                self._set_iio_attr(self.name, "raw1", True, str(int(val)))

        @property
        def volt0(self):
            """ LTC2688 channel toggle state 0 voltage value """
            return ((self.raw0 + self.offset) * self.scale) * (ltc2688.vref / 4.096)

        @volt0.setter
        def volt0(self, val):
            """ LTC2688 channel toggle state 0 voltage value setter """
            self.raw0 = int(((val / self.scale) - self.offset) * (ltc2688.vref / 4.096))

        @property
        def volt1(self):
            """ LTC2688 channel toggle state 1 voltage value """
            return ((self.raw1 + self.offset) * self.scale) * (ltc2688.vref / 4.096)

        @volt1.setter
        def volt1(self, val):
            """ LTC2688 channel toggle state 1 voltage value setter """
            self.raw1 = int(((val / self.scale) - self.offset) * (ltc2688.vref / 4.096))

    class _channel_sw_toggle(_channel_toggle):
        def __init__(self, ctrl, channel_name):
            super().__init__(ctrl, channel_name)

        @property
        def toggle_state(self):
            """ LTC2688 SW toggle enable flag """
            return self._get_iio_attr(self.name, "symbol", True, self._ctrl)

        @toggle_state.setter
        def toggle_state(self, val):
            """ LTC2688 SW toggle enable flag setter """
            self._set_iio_attr(self.name, "symbol", True, str(int(val)))
